<?php

namespace Drupal\improvements_paragraphs\Plugin\paragraphs\Behavior;

use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element\Checkboxes;
use Drupal\Core\Render\Markup;
use Drupal\druhels\ArrayHelper;
use Drupal\paragraphs\Entity\Paragraph;
use Drupal\paragraphs\ParagraphInterface;
use Drupal\paragraphs\ParagraphsBehaviorBase;

/**
 * @ParagraphsBehavior(
 *   id = "style_selector",
 *   label = @Translation("Style selector"),
 *   description = @Translation("Allows to select paragraph style from predefined list."),
 *   weight = 103,
 * )
 */
class StyleSelectorParagraphBehavior extends ParagraphsBehaviorBase {

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'styles' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $default_value_array = [];
    foreach ($this->configuration['styles'] as $style_id => $style) {
      $row = $style_id . '|' . $style['label'] . '|' . $style['class'];

      // Additional options
      $i = 0;
      while (isset($style['option' . $i])) {
        $row .= '|' . $style['option' . $i];
        $i++;
      }

      $default_value_array[] = $row;
    }

    $form['styles'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Styles'),
      '#description' => $this->t('Format: @format', ['@format' => Markup::create('<code>style_id|Style name|css-class</code>')]),
      '#default_value' => implode("\n", $default_value_array),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $styles = [];
    $rows = ArrayHelper::getTextLines($form_state->getValue('styles'));

    foreach ($rows as $row) {
      $row_array = explode('|', $row);
      $styles[$row_array[0]] = [
        'label' => $row_array[1],
        'class' => $row_array[2] ?? '',
      ];

      $additional_options = array_slice($row_array, 3);
      foreach ($additional_options as $key => $additional_option) {
        $styles[$row_array[0]]['option' . $key] = $additional_option;
      }
    }

    $this->configuration['styles'] = $styles;
  }

  /**
   * {@inheritdoc}
   */
  public function buildBehaviorForm(ParagraphInterface $paragraph, array &$form, FormStateInterface $form_state): array {
    $styles_options = [];
    foreach ($this->configuration['styles'] as $style_id => $style) {
      $styles_options[$style_id] = $style['label'];
    }

    $form['styles'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Style'),
      '#options' => $styles_options,
      '#default_value' => $paragraph->getBehaviorSetting($this->pluginId, 'styles'),
    ];

    $form['#weight'] = $this->pluginDefinition['weight'];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitBehaviorForm(ParagraphInterface $paragraph, array &$form, FormStateInterface $form_state): void {
    $filtered_values = $this->filterBehaviorFormSubmitValues($paragraph, $form, $form_state);
    $styles = !empty($filtered_values['styles']) ? Checkboxes::getCheckedCheckboxes($filtered_values['styles']) : [];
    $paragraph->setBehaviorSettings($this->pluginId, ['styles' => $styles]);
  }

  /**
   * {@inheritDoc}
   */
  public function view(array &$build, Paragraph $paragraph, EntityViewDisplayInterface $display, $view_mode): void {
    $all_paragraph_behavior_settings = $paragraph->getAllBehaviorSettings(); // Not used getBehaviorSetting() for increase performance
    $selected_styles = $all_paragraph_behavior_settings[$this->pluginId]['styles'] ?? [];
    $css_classes = [];

    foreach ($selected_styles as $style_id) {
      if (isset($this->configuration['styles'][$style_id])) {
        $css_classes[] = $this->configuration['styles'][$style_id]['class'];
      }
    }

    if ($css_classes) {
      $build['#attributes']['class'] = array_merge($build['#attributes']['class'] ?? [], $css_classes);
    }
  }

}
